"""
A displacement controlled triaxial test of an elastic material in 2D
"""

# Importing the essentials
from fenics import *
import matplotlib.pyplot as plt
import numpy as np

# Some useful parameters
i = Index()
j = Index()
k = Index()
l = Index()
# Units are in SI
# Geometry
Lx = 0.1	# width
Ly = 0.2	# height
nx = 5		# number of elements in the x-direction
ny = 10		# number of elements in the y-direction

# Material parameters
nu = 0.20
E  = 10000000.0
lamda = Constant(E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu)))
mu    = Constant(E / (2.0 * (1.0 + nu)))
rho   =  0.0#50000.0
g     = 10.0 

# Preparing the mesh
mesh = RectangleMesh(Point(0,0), Point(Lx,Ly), nx, ny)
plot(mesh, title="Mesh")
plt.show()

# Defining elements
# Second order to avoid locking
V = VectorFunctionSpace(mesh, 'Lagrange', 2)
#Define test functions ()
v = TestFunction(V)
#Define trial functions (unknown displacements)
u = TrialFunction(V)

# Define Dirichlet boundary conditions
# Lower boundary, zero vertical displacement
tol = 1E-14
def lower_boundary(x, on_boundary):
    return on_boundary and x[1] < tol

bcd = DirichletBC(V.sub(1), Constant( 0.001), lower_boundary)

# Upper boundary, given vertical displacement
def upper_boundary(x, on_boundary):
    return on_boundary and x[1] > Ly-tol
	
bcu = DirichletBC(V.sub(1), Constant(-0.001), upper_boundary)
	
class LeftBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] < tol

# right boundary
class RightBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0]> Lx - tol	

left_boundary = LeftBoundary()        
right_boundary = RightBoundary()

bcl = DirichletBC(V.sub(0), Constant( 0.0005), left_boundary)
bcr = DirichletBC(V.sub(0), Constant(-0.0005), right_boundary)

bc = [bcd, bcu, bcl, bcr]

# Strain
def epsilon(u):
    e =-0.5*(nabla_grad(u) + nabla_grad(u).T)
    #e =-nabla_grad(u)
    return as_tensor([[e[0, 0], e[0, 1]],
                      [e[1, 0], e[1, 1]]])

# Stiffness tensor					  
def dsde():
    stiffness = np.zeros((d,d,d,d))
    for i in range(0,d):
         for j in range(0,d):
             stiffness[i,i,j,j] +=lamda
             stiffness[i,j,i,j] +=mu
             stiffness[i,j,j,i] +=mu
    return as_tensor(stiffness)
	
# Stress tensor
def sigma(u):
    eps=epsilon(u)
    stiffness =dsde()
    sigma = as_tensor(stiffness[i,j,k,l]*eps[k,l],(i,j))
    return as_tensor(sigma)

# Define variational problem
u = TrialFunction(V)
d = u.geometric_dimension()  # space dimension
v = TestFunction(V)
f = Constant(( 0, -rho*g))
T = Constant(( 0, 0))
Ti= Constant(( 0, 0))

# Assembling
a = inner(sigma(u), nabla_grad(v))*dx
L =-dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2)

q = Function(V)
solve(a == L, q, bc)

u=q

fmt='%1.8f'

U = FunctionSpace(mesh, 'P', 1)
# Compute magnitude of displacement
u_magnitude = sqrt(dot(u, u))
u_magnitude = project(u_magnitude, U)
p=plot(u_magnitude, title='Displacement magnitude')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(u[1], title='Displacement vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

# Plot stress
s = sigma(u) # all stress
V = FunctionSpace(mesh, 'P', 1)
s0 = s[0,0]
s1 = s[1,1]

p=plot(s0, title='Stress horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(s1, title='Stress vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

print('min/max u:',
      u_magnitude.vector().get_local().min(),
      u_magnitude.vector().get_local().max())

# this integrates over the domain
energy = 0.5*inner(sigma(u), epsilon(u))*dx
E = assemble(energy)	

print('energy:',
      E)
	  



