"""
A displacement controlled triaxial test of an elastic material in 2D
"""

# Importing the essentials
from fenics import *
import matplotlib.pyplot as plt
import numpy as np

# Some useful parameters
i = Index()
j = Index()
k = Index()
l = Index()
# Units are in SI
# Geometry
Lx = 0.1	# width
Ly = 0.2	# height
nx = 5		# number of elements in the x-direction
ny = 10		# number of elements in the y-direction

# For the Newton-Raphson
niter_max = 20     # maximum number of iterations  
prec      = 1e-8   # precision

# Material parameters
nu = 0.2
E  = 10000000.0
lamda = Constant(E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu)))
mu    = Constant(E / (2.0 * (1.0 + nu)))
rho   =  0.0#50000.0
g     = 10.0 

# Preparing the mesh
mesh = RectangleMesh(Point(0,0), Point(Lx,Ly), nx, ny)
plot(mesh, title="Mesh")
plt.show()

# Defining elements
# Second order to avoid locking
V = VectorFunctionSpace(mesh, 'Lagrange', 2)
#Define test functions ()
v = TestFunction(V)
#Define trial functions (unknown displacements)
u = TrialFunction(V)

# Define Dirichlet boundary conditions
# Lower boundary, zero vertical displacement
tol = 1E-14
def lower_boundary(x, on_boundary):
    return on_boundary and x[1] < tol
	
class LeftBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] < tol

# right boundary
class RightBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] > Lx - tol	
		
class UpperBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[1] > Ly - tol

left_boundary = LeftBoundary()        
right_boundary = RightBoundary()
upper_boundary = UpperBoundary()

boundary_parts = MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
boundary_parts.set_all(0)
left_boundary.mark(boundary_parts, 1)
right_boundary.mark(boundary_parts, 2)
upper_boundary.mark(boundary_parts, 3)
ds = ds(subdomain_data = boundary_parts)


bcd = DirichletBC(V.sub(1), Constant(0.000), lower_boundary)


bc = [bcd]

# Strain
def epsilon(u):
    e =-0.5*(nabla_grad(u) + nabla_grad(u).T)
    #e =-nabla_grad(u)
    return as_tensor([[e[0, 0], e[0, 1]],
                      [e[1, 0], e[1, 1]]])

# Stiffness tensor					  
def dsde():
    stiffness = np.zeros((d,d,d,d))
    for i in range(0,d):
         for j in range(0,d):
             stiffness[i,i,j,j] +=lamda
             stiffness[i,j,i,j] +=mu
             stiffness[i,j,j,i] +=mu
    return as_tensor(stiffness)
	
# Stress tensor
def sigma(u):
    eps=epsilon(u)
    stiffness =dsde()
    sigma = as_tensor(stiffness[i,j,k,l]*eps[k,l],(i,j))
    return as_tensor(sigma)
	
# Newton-Raphson matrix
def NR(u):
     stiffness =dsde()
     stf=-as_tensor(stiffness[(i,j,k,l)]*nabla_grad(u)[(k,l)],(i,j))
     NR = inner(stf, nabla_grad(v))*dx
     return NR

# Newton-Raphson residual
def res(sol):
    res = -inner(sigma(sol), nabla_grad(v))*dx + dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2)	
    return res	

# Define variational problem
u = TrialFunction(V)
trialf = TrialFunction(V)
d = u.geometric_dimension()  # space dimension
v = TestFunction(V)
f = Constant(( 0, -rho*g))
T = Constant((-50000., 0))
Ti= Constant(( 50000., 0))
Si= Constant(( 0, 50000.))

# Assembling
a = inner(sigma(u), nabla_grad(v))*dx
L = dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2) + dot(Si,v)*ds(3)

u0= Function(V)
solve(a == L, u0, bc)

fmt='%1.8f'

p=plot(u0[1], title='Displacement vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(u0[0], title='Displacement horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

# Plot stress
s = sigma(u0) # all stress
s0 = s[0,0]
s1 = s[1,1]

p=plot(s0, title='Stress horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(s1, title='Stress vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()


# Units are in SI
# Concerning the time
Tf = 1.0            # final time
num_steps = 50   	# number of time steps
dt = Tf / num_steps # time step size

# useful variables
U  = Function(V, name="Total displacement")
Du = Function(V, name="Current increment")
dU = Function(V, name="Iteration correction")
	
# Time-stepping
t = 0

Uf= FunctionSpace(mesh, 'P', 1)
dd = project(u0[1], Uf)

# Displacement increment
dudisp=-0.002*dt/Tf
# Initializing
U = u0

for n in range(num_steps):
# Update current time
    t += dt

# Incremental boundary conditions	
    bcu = DirichletBC(V.sub(1), Constant(dudisp), upper_boundary)
    bcs = [bcd, bcu]	
# Starting the Newton-Raphson
    niter = 0   # initialize iterations
    nRes  = 1.  # initialize residue
    Du.interpolate(Constant((0, 0)))

    while nRes > prec and niter < niter_max:
        res = -inner(sigma(U+Du), nabla_grad(v))*dx + dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2)		
        solve(NR(u)==res, dU, bcs)
        Du.assign(Du+dU)	
        res = -inner(sigma(U+Du), nabla_grad(v))*dx + dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2)			
        Res = assemble(res)
        for bc in bcs:
            bc.homogenize()
            bc.apply(Res)
        nRes = Res.norm("l2")
        ndu  = dU.vector().norm('l2')
        print("    Residual:", nRes , "Increment:", ndu ,"Iteration:", niter)
#        nRes = 0
        niter += 1
    U.assign(U+Du)	
u = U

fmt='%1.8f'

U = FunctionSpace(mesh, 'P', 1)
# Compute magnitude of displacement
u_magnitude = sqrt(dot(u, u))
u_magnitude = project(u_magnitude, Uf)
p=plot(u_magnitude, title='Displacement magnitude')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(u[1], title='Displacement vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(u[0], title='Displacement horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

# Plot stress
s = sigma(u) # all stress
V = FunctionSpace(mesh, 'P', 1)
s0 = s[0,0]
s1 = s[1,1]

p=plot(s0, title='Stress horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(s1, title='Stress vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

print('min/max u:',
      u_magnitude.vector().get_local().min(),
      u_magnitude.vector().get_local().max())

# this integrates over the domain
energy = 0.5*inner(sigma(u), epsilon(u))*dx
E = assemble(energy)	

print('energy:',
      E)
	  



