#Modified from the chapter on virtual power
#Import FEniCS library
from dolfin import * 
#Parameters
h=2.		#height of the layer
s12bc=.5	#applied shear stress at boundary
s22bc=.25	#applied normal stress at boundary
m3bc=.0		#applied Cosserat moment at boundary
Eel=1000.	#Young modulus
nuel=0.		#Poisson ratio
eta1=.8		#Cosserat coefficient for Gc
eta3=5./2.	#Cosserat coefficient for Mc
lc=h/3.		#Cosserat length

ny=10  		#number of elements-FE discretization

Gel=Eel/(2.*(1+nuel))
Mel=Eel*(1.-nuel)/((1.-2.*nuel)*(1.+nuel))
Gcel=eta1*Gel
Mcel=2*eta3*Gel*lc**2

#Generate mesh
mesh=IntervalMesh(ny,0.,h/2.)
#Define element topology (1D)
cell=interval
#Defines a Lagrangian FE of degree 2 for the displacements
elementdisp=VectorElement("Lagrange",cell,degree=2,dim=2)
#Defines a Lagrangian FE of degree 1 for the rotations
elementrot=FiniteElement("Lagrange",cell,degree=1)
#Creates a mixed element
element=elementdisp*elementrot
#Assign the element to the mesh
V=FunctionSpace(mesh,element)
#Define test functions (virtual velocities)
v=TestFunction(V)
#Define trial functions (unknown displacements and Cos. rotation)
u=TrialFunction(V)

#Store the solution to sol
sol=Function(V)

#Define boundary conditions
def middle(x,on_boundary):
	return on_boundary and near(x[0],0.)
bc=DirichletBC(V,(0.,0.,0.),middle)
#Define traction vector
timui=Constant((s22bc,s12bc,m3bc))

#Define internal virtual power
Pint=(
	Mel*Dx(u[0],0)*Dx(v[0],0)+#tau22*gamma22
	((Gel+Gcel)*Dx(u[1],0)+2*Gcel*u[2])*(Dx(v[1],0)+v[2])+#tau12*gamma12
	((Gel-Gcel)*Dx(u[1],0)-2*Gcel*u[2])*(-v[2])+#tau21*gamma21
	Mcel*Dx(u[2],0)*Dx(v[2],0)#m32*k32
	)*dx
#Define internal virtual power
Pext=dot(timui,v)*ds

#Solve the problem (does the FE formulation, matrix assembly and linearsolve)
solve(Pint==Pext,sol,bc)

#Plot solution
import matplotlib.pyplot as plt
plot(sol[1],label='$u_x$')
plot(sol[0],label='$u_y$')
plot(-sol[2],label='$-\omega_z$')
plt.xlabel("$x_2$")
plt.legend(loc='upper left')
plt.ylim(0,.001)
font={'size':18}
plt.rc('font',**font)
plt.show()

# evaluating the stored energy
energys= 0.5*(
 Mel*Dx(sol[0],0)*Dx(sol[0],0)+#tau11*gamma11	
 ((Gel+Gcel)*Dx(sol[1],0)+2*Gcel*sol[2])*(Dx(sol[1],0)+sol[2])+#tau12*gamma12
 ((Gel-Gcel)*Dx(sol[1],0)-2*Gcel*sol[2])*(-sol[2])#tau21*gamma21	
 )*dx
Es = assemble(energys)	

energym= 0.5*(
 Mcel*Dx(sol[2],0)*Dx(sol[2],0)	
 )*dx
Em = assemble(energym)	

E  = Em + Es	

print('energy stress:',
      Es,
      'energy moment:',
	  Em,
	  'energy total:',
	  E)