"""
This is me trying to figure this out
Going for a triaxial test of an elastic material in 2D time-dependent with stiffness
linear and anisotropic
"""

# Importing the essentials
from fenics import *
import matplotlib.pyplot as plt
import numpy as np

# Geometry
Lx = 0.1	# width
Ly = 0.2	# height
nx = 50		# number of elements in the x-direction
ny = 100		# number of elements in the y-direction

# Material parameters
nu = 0.20
G  = 1000000.0
C  = 5.
phs= 0.5
pf = phs*G*C/(16*(1.-nu))

# Anisotropy parameters
cc = 0.0
dd = 0.0
psi= 0.*pi/6. 

# Density and gravity
rho   = 0.0
g     = 10.0 

# Preparing the mesh
mesh = RectangleMesh(Point(0,0), Point(Lx,Ly), nx, ny)
#plot(mesh, title="Mesh")
#plt.show()

# Defining elements
# Second order to avoid locking
V = VectorFunctionSpace(mesh, 'Lagrange', 2)
#Define test functions ()
v=TestFunction(V)
#Define trial functions (unknown displacements)
u=TrialFunction(V)

# Define Dirichlet boundary conditions
# u_y on lower boundary
tol = 1E-14
def lower_boundary(x, on_boundary):
    return on_boundary and x[1] < tol

bcd = DirichletBC(V.sub(1), Constant(0.0), lower_boundary)

# Clamping a point
particular_point = "near(x[0], 0.05) && near(x[1], 0.0)"
bcm=DirichletBC(V,Constant((0.0,0.0)),particular_point,"pointwise")

# Defining stress boundary conditions
# left boundary
class LeftBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] < tol

# right boundary
class RightBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0]> Lx - tol	

class UpperBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[1] > Ly - tol

left_boundary = LeftBoundary()        
right_boundary = RightBoundary()
upper_boundary = UpperBoundary()

boundary_parts = MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
boundary_parts.set_all(0)
left_boundary.mark(boundary_parts, 1)
right_boundary.mark(boundary_parts, 2)
upper_boundary.mark(boundary_parts, 3)
ds = ds(subdomain_data = boundary_parts)

# Define strain and stress
def epsilon(u):
    e = 0.5*(nabla_grad(u) + nabla_grad(u).T)
    return as_tensor([[e[0, 0], e[0, 1]],
                      [e[1, 0], e[1, 1]]])

# Stiffness tensor					  
def dsde():
    stiffness = np.zeros((d,d,d,d))
    stiffness[0,0,0,0]=pf*(2.*(4-nu)-4*cc*cos(2.*psi)-dd*nu*cos(4.*psi))
    stiffness[0,0,0,1]=pf*(2*cc*(2.-nu)*sin(2.*psi)+dd*nu*sin(4.*psi))
    stiffness[0,0,1,0]=pf*(2*cc*nu*sin(2.*psi)+dd*nu*sin(4.*psi))
    stiffness[0,0,1,1]=pf*(2*nu+dd*nu*cos(4.*psi))
	
    stiffness[0,1,0,0]=pf*(2.*cc*(2.-nu)*sin(2.*psi)+dd*nu*sin(4.*psi))
    stiffness[0,1,0,1]=pf*(8.-6.*nu+4.*cc*(1.-nu)*cos(2.*psi)+dd*nu*cos(4.*psi))
    stiffness[0,1,1,0]=pf*(2.*nu+dd*nu*cos(4.*psi))
    stiffness[0,1,1,1]=pf*(2.*cc*nu*sin(2.*psi)-dd*nu*sin(4.*psi))
	
    stiffness[1,0,0,0]=pf*(2.*cc*nu*sin(2.*psi)+dd*nu*sin(4.*psi))
    stiffness[1,0,0,1]=pf*(2.*nu+dd*nu*cos(4.*psi))
    stiffness[1,0,1,0]=pf*(8.-6.*nu-4.*cc*(1.-nu)*cos(2.*psi)+dd*nu*cos(4.*psi))
    stiffness[1,0,1,1]=pf*(2.*cc*(2.-nu)*sin(2.*psi)-dd*nu*sin(4.*psi))
	
    stiffness[1,1,0,0]=pf*(2*nu+dd*nu*cos(4.*psi))
    stiffness[1,1,0,1]=pf*(2*cc*nu*sin(2.*psi)-dd*nu*sin(4.*psi))
    stiffness[1,1,1,0]=pf*(2*cc*(2.-nu)*sin(2.*psi)-dd*nu*sin(4.*psi))
    stiffness[1,1,1,1]=pf*(2.*(4-nu)+4*cc*cos(2.*psi)-dd*nu*cos(4.*psi))	

##   Forcing symmetry	
#    stiffness[0,0,0,1]=(stiffness[0,0,0,1]+stiffness[0,0,1,0])/2.0
#    stiffness[0,0,1,0]=stiffness[0,0,0,1]
	
#    stiffness[0,1,0,1]=(stiffness[0,1,0,1]+stiffness[0,1,1,0])/2.0
#    stiffness[0,1,1,0]=stiffness[0,1,0,1]

#    stiffness[1,0,0,1]=(stiffness[1,0,0,1]+stiffness[1,0,1,0])/2.0
#    stiffness[1,0,1,0]=stiffness[1,0,0,1]

#    stiffness[1,1,0,1]=(stiffness[1,1,0,1]+stiffness[1,1,1,0])/2.0
#    stiffness[1,1,1,0]=stiffness[1,1,0,1]

    stiffness[0,1,:,:] =(stiffness[0,1,:,:]+stiffness[1,0,:,:])/2.0
    stiffness[1,0,:,:] = stiffness[0,1,:,:] 
    return as_tensor(stiffness)
	
# Stress tensor
def sigma(u):
#    print('dimensions:',
#         dsde().shape,
#         epsilon(u).shape)
#    sigma = np.tensordot(dsde(),epsilon(u),axes=([2,0],[3,1]))
    eps=epsilon(u)
    stiffness =dsde()
    sigma = as_tensor(stiffness[i,j,k,l]*eps[k,l],(i,j))
#    sigma = lamda*tr(eps)*Identity(d) + 2*mu*eps	
    return as_tensor(sigma)
	
#f = Expression("10*exp(-(pow(x[0] - 0.5, 2) + pow(x[1] - 0.5, 2)) / 0.02)")

# Define variational problem
u = TrialFunction(V)
d = u.geometric_dimension()  # space dimension
v = TestFunction(V)
f = Constant((0, -rho*g))
T = Constant(( 100000., 0.))
Ti= Constant((-100000., 0.))
Si= Constant(( 0.,-150000.))

# Assembling
a = inner(sigma(u), nabla_grad(v))*dx
L = dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2)+ dot(Si,v)*ds(3)	
## Update the Dirichlet boundary condition	
#udisp=-0.001
#bcu = DirichletBC(V.sub(1), Constant(udisp), upper_boundary)
#bc = [bcd, bcm, bcu]
bc = [bcd, bcm]
# Compute solution
q = Function(V)
solve(a == L, q, bc)

u=q

fmt='%1.8f'

# Plot solution
p=plot(u, title='Displacement', mode='displacement')
plt.colorbar(p,format=fmt)
# Hold plot
plt.show()

p=plot(u[0], title='Displacement horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(u[1], title='Displacement vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(epsilon(u)[(1,1)], title='Strain vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()
p=plot(epsilon(u)[(0,0)], title='Strain horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

# Plot stress
s = sigma(u) # all stress
V = FunctionSpace(mesh, 'P', 1)
s0 = s[(0,0)]
s1 = s[(1,1)]
sa = s[(1,0)]-s[(0,1)]
press = (s0+s1)/2.0
strq  =sqrt((s1-s0)**2/4+s[(1,0)]**2)

p=plot(press, title='Mean stress')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(strq, title='Deviatoric stress')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(s1, title='Vertical stress')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(s0, title='Horizontal stress')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

# this integrates over the domain
energy = 0.5*inner(sigma(u), grad(u))*dx
E = assemble(energy)	

print('energy:',
      E)
	  


