"""
A displacement controlled triaxial test of an elastic material in 2D
"""

# Importing the essentials
from dolfin import *
import matplotlib.pyplot as plt
import numpy as np
#parameters.form_compiler.quadrature_degree = 2 # or some higher number
#parameters["form_compiler"]["representation"] = "quadrature"

set_log_level(INFO)

# Some useful parameters
i = Index()
j = Index()
k = Index()
l = Index()
ii= Index()
ji= Index()
ki= Index()
li= Index()

fmt='%1.8f'
# Units are in SI
# Geometry
Lx = 0.1	# width
Ly = 0.2	# height
nx = 5		# number of elements in the x-direction
ny = 10		# number of elements in the y-direction

# For the Newton-Raphson
niter_max = 20     # maximum number of iterations  
prec      = 1e-8   # precision

# Material parameters
# Elastic
nu = 0.2
E  = 1000000.0
lamda = Constant(E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu)))
mu    = Constant(E / (2.0 * (1.0 + nu)))

# Plastic
cu =5.
phi=pi/6.
alpha=0.0
af =Constant(sin(phi))
ag =Constant(sin(phi+alpha))
mult = Constant((mu+(mu+lamda)*sin(phi)*sin(phi+alpha)))
# Physical
rho   =  0.0#50000.0
g     = 10.0 

# Preparing the mesh
mesh = RectangleMesh(Point(0,0), Point(Lx,Ly), nx, ny)
#plot(mesh, title="Mesh")
#plt.show()

# We will be needing elastic and plastic strains and the stresses
# and we prefer to save those at the quadrature points
# Create function space
deg_u = 2
ns = deg_u-1

V = VectorFunctionSpace(mesh, "CG", deg_u)
We = TensorElement("Quadrature", mesh.ufl_cell(), degree=ns,shape=(2,2),quad_scheme='default')
W = FunctionSpace(mesh,We)
W0e = FiniteElement("Quadrature", mesh.ufl_cell(), degree=ns,quad_scheme='default')
W0 = FunctionSpace(mesh, W0e)
P0 = FunctionSpace(mesh,"DG",0)



# Define Dirichlet boundary conditions
# Lower boundary, zero vertical displacement
tol = 1E-14
def lower_boundary(x, on_boundary):
    return on_boundary and x[1] < tol

# Clamping a point
particular_point = "near(x[0], 0.05) && near(x[1], 0.0)"
	
class LeftBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] < tol

# right boundary
class RightBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] > Lx - tol	
		
class UpperBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[1] > Ly - tol

left_boundary = LeftBoundary()        
right_boundary = RightBoundary()
upper_boundary = UpperBoundary()

boundary_parts = MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
boundary_parts.set_all(0)
left_boundary.mark(boundary_parts, 1)
right_boundary.mark(boundary_parts, 2)
upper_boundary.mark(boundary_parts, 3)
ds = ds(subdomain_data = boundary_parts)

metadata = {"quadrature_degree": ns, "quadrature_scheme": "default"}
dxm = dx(metadata=metadata)

bcd = DirichletBC(V.sub(1), Constant(0.000), lower_boundary)
bcm=DirichletBC(V,Constant((0.0,0.0)),particular_point,"pointwise")
bc = [bcd, bcm]

# Strain
def epsilon(u):
    e =-0.5*(nabla_grad(u) + nabla_grad(u).T)
    #e =-nabla_grad(u)
    return as_tensor([[e[0, 0], e[0, 1]],
                      [e[1, 0], e[1, 1]]])

# Stiffness tensor					  
def dsde():
    stiffness = np.zeros((d,d,d,d))
    for i in range(0,d):
         for j in range(0,d):
             stiffness[i,i,j,j] +=lamda
             stiffness[i,j,i,j] +=mu
             stiffness[i,j,j,i] +=mu
    return as_tensor(stiffness)
	
# Stress tensor
def sigma_el(eps):
    stiffness =dsde()
    sigma_el = as_tensor(stiffness[i,j,k,l]*eps[k,l],(i,j))
    return as_tensor(sigma_el)

def sigma_pl(eps):
    stiffness =dsde()
    sigma_pl = as_tensor(stiffness[i,j,k,l]*eps[k,l],(i,j))
    return as_tensor(sigma_pl)	
	
# Newton-Raphson matrix
# Elastoplastic iteration
def NR(u,sig):
    stiffness =dsde()
    s=dev(sig)
    q=sqrt(inner(s,s)/2.)
    p = tr(sig)/2.0
    f_y = q - af*p - cu	
    df_y = (s/q-af*Identity(d))/2.
    dg_y = (s/q-ag*Identity(d))/2.
    H = df_y[i,j]*stiffness[i,j,k,l]*dg_y[k,l]	
    lpl = project(pone(f_y)/H,Uf)    	
    stf_pl = lpl*as_tensor(dg_y[(ii,ji)]*stiffness[(i,j,ii,ji)]*stiffness[(ki,li,k,l)]*df_y[(ki,li)],(i,j,k,l))
    stf_pl =-as_tensor(stiffness-stf_pl)
    stf= as_tensor(stf_pl[(i,j,k,l)]*nabla_grad(u)[(k,l)],(i,j))
#    stf=-as_tensor(stiffness[(i,j,k,l)]*nabla_grad(u)[(k,l)],(i,j))
    NR = inner(stf,nabla_grad(v))*dx
    return NR

# Macaulay bracket
ppos = lambda x: (x+abs(x))/2.

# Heaviside
pone = lambda x: (1+x/(abs(x)+1e-8))/2.

def dev(a):
    return a -1/2.*tr(a)*Identity(2)
	
def mapping(sig0, deps):
    sig_el = as_tensor(sig0) + sigma_el(epsilon(deps))		
    s=dev(sig_el)
    q=sqrt(inner(s,s)/2.)	
    p = tr(sig_el)/2.0
    f_y = q - af*p - cu
#    pf=plot(f_y, title='yf')
# add a colorbar
#    plt.colorbar(pf,format=fmt);
# Hold plot
#    plt.show()	
    lpl = ppos(f_y)/mult
    dg_y = (s/q-ag*Identity(d))/2.
    #df_y = (af*Identity(d))/2.	
    de_plas= project(lpl*dg_y,W)
    de_elas= epsilon(deps)-as_tensor(de_plas)
    sig_new= sig0 + sigma_el(de_elas)	
    return  as_tensor([[sig_new[0, 0], sig_new[0, 1]],
            [sig_new[1, 0], sig_new[1, 1]]]), \
            as_tensor([[de_plas[0, 0], de_plas[0, 1]],
            [de_plas[1, 0], de_plas[1, 1]]])

def local_project(v,S,uu=None):
    dv = TrialFunction(S)
    v_ = TestFunction(S)
    a_proj = inner(dv,v_)*dx(metadata=metadata)
    b_proj = inner(v,v_)*dx(metadata=metadata)
    solver = LocalSolver(a_proj,b_proj)
    solver.factorize()
    if uu is None:
        uu = Function(S)
        solver.solve_local_rhs(uu)
        return uu
    else:
        solver.solve_local_rhs(uu)
        return

def as_2D_tensor(X):
    return as_tensor([[X[0], X[2]],
                      [X[3], X[1]]])
# Define variational problem
u = TrialFunction(V)
d = u.geometric_dimension()  # space dimension
v = TestFunction(V)
f = Constant(( 0, -rho*g))
T = Constant((-50000., 0))
Ti= Constant(( 50000., 0))
Si= Constant(( 0, 50000.))
Si= Constant(( 0, 145000.))
# Assembling
a = inner(sigma_el(epsilon(u)), nabla_grad(v))*dx
L = dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2) + dot(Si,v)*ds(3)

u0= Function(V)
solve(a == L, u0, bc)

# Saving the boundary information for the loading step
Uf= FunctionSpace(mesh, 'CG', 2)
dd = project(u0[1], Uf)



# Done with elastic preloading, proceeding to triaxial compression
# Units are in SI
# Concerning the time
Tf = 1.0            # final time
num_steps = 50   	# number of time steps
dt = Tf / num_steps # time step size

# useful variables
# vectors at nodes
U  = Function(V, name="Total displacement")
Du = Function(V, name="Current increment")
dU = Function(V, name="Iteration correction")
# tensors at quadrature points
W = TensorFunctionSpace(mesh, "CG", 2)
sig     = Function(W, name="Current stress")
sig_old = Function(W, name="Stress at previous step")
e_plas  = Function(W,  name="Plastic strain")
de_plas = Function(W,  name="Plastic strain increment")

sig_ = Constant(as_tensor([[0,0],[0,0]]))
de_plas_ = Constant(as_tensor([[0,0],[0,0]]))
# scalars at quadrature points
beta   = Function(W0, name="Placeholder 1 for evolving plasticity")
alph   = Function(W0, name="Placeholder 2 for evolving plasticity")
# Test and trial functions
u = TrialFunction(V)
v = TestFunction(V)
uu= Function(V)

# Time-stepping
t = 0
# Displacement increment
dudisp=-0.002*dt/Tf

bcu = DirichletBC(V.sub(1), Constant(dudisp), upper_boundary)
bcu0= DirichletBC(V.sub(1), Constant(0.0), upper_boundary)
# Initializing
U = u0

sig_elas =sigma_el(epsilon(U))
sig = sig_elas
sig_old = sig_elas
p=plot(sig[(1,1)], title='checking project yy')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()
p=plot(sig[(0,0)], title='checking project xx')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()
e_plas.interpolate(de_plas)


res1= np.zeros(num_steps)
res2= np.zeros(num_steps)
res3= np.zeros(num_steps)
res4= np.zeros(num_steps)

for n in range(num_steps):
# Update current time
    t += dt

# Incremental boundary conditions	
    bcs = [bcd, bcm, bcu]	
# Starting the Newton-Raphson
    niter = 0   # initialize iterations
    nRes  = 1.  # initialize residue
    Du.interpolate(Constant((0, 0)))
    print("N:", n ,"Iteration:", niter,"Checkpoint:", 0)	
    while nRes > prec and niter < niter_max:
        res = -inner(sig, nabla_grad(v))*dx + dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2)	
        solve(NR(u,sig)==res, dU, bcs)
#        p=plot(dU[1], title='duy')
# add a colorbar
#        plt.colorbar(p,format=fmt);
# Hold plot
#        plt.show()		
#        print("N:", n ,"Iteration:", niter,"Checkpoint:", 1)	
        Du.assign(Du+dU)					
        sig_,de_plas_= mapping(sig_old,Du)	
#        print("N:", n ,"Iteration:", niter,"Checkpoint:", 3)	
        sig = project(sig_,W) 
        de_plas = project(de_plas_,W) 			
#        local_project(sig_, W, sig)
#        local_project(de_plas_, W, de_plas)
# Incremental boundary conditions	
        bcs = [bcd, bcm, bcu0]
        res = -inner(sig, nabla_grad(v))*dx + dot(f, v)*dx + dot(T, v)*ds(1) + dot(Ti,v)*ds(2)	
#        print("N:", n ,"Iteration:", niter,"Checkpoint:", 4)		
        Res = assemble(res)
#        print("N:", n ,"Iteration:", niter,"Checkpoint:", 5)				
        for bc in bcs:
            bc.homogenize()
            bc.apply(Res)
#        print("N:", n ,"Iteration:", niter,"Checkpoint:", 6)			
        nRes = Res.norm("l2")
        ndu  = dU.vector().norm('l2')

#        print("N:", n ,"Iteration:", niter,"Checkpoint:", 7)		
        print("    Residual:", nRes , "Increment:", ndu ,"Iteration:", niter)
        niter += 1		
    U.assign(U+Du)	
    sig_old=sig
    e_plas+=de_plas	
    res_tmp = U[1]*ds(3)
    res1[n]   =assemble(res_tmp)/Lx		
    res_tmp = sig[(1,1)]*ds(3)
    res2[n]   =assemble(res_tmp)/Lx 
    res_tmp = sig[(0,0)]*ds(1)
    res3[n]   =assemble(res_tmp)/Ly
    res_tmp = epsilon(U)[0,0]*dx
    res4[n]   =assemble(res_tmp)/(Lx*Ly)	
	
#    p=plot(sig[(1,1)], title='Stress project yy')
# add a colorbar
#    plt.colorbar(p,format=fmt);
# Hold plot
#    plt.show()		
#    p=plot(sig[(0,0)], title='Stress project xx')
# add a colorbar
#    plt.colorbar(p,format=fmt);
# Hold plot
#    plt.show()	
u = U

fmt='%1.8f'


plt.plot(res1,res2, 'ko',label='yy')
plt.xlabel("$c$")
plt.ylabel("$E_{tot}$")
plt.legend(loc='upper left')
#plt.ylim(0,.0001)
font={'size':24}
plt.rc('font',**font)
plt.show()

plt.plot(res4,res2, 'ko',label='yy')
plt.xlabel("$c$")
plt.ylabel("$E_{tot}$")
plt.legend(loc='upper left')
#plt.ylim(0,.0001)
font={'size':24}
plt.rc('font',**font)
plt.show()

plt.plot(res1,res3, 'ko',label='xx')
plt.xlabel("$c$")
plt.ylabel("$E_{tot}$")
plt.legend(loc='upper left')
#plt.ylim(0,.0001)
font={'size':24}
plt.rc('font',**font)
plt.show()

plt.plot(res2,res3, 'ko',label='xy')
plt.xlabel("$c$")
plt.ylabel("$E_{tot}$")
plt.legend(loc='upper left')
#plt.ylim(0,.0001)
font={'size':24}
plt.rc('font',**font)
plt.show()
U = FunctionSpace(mesh, 'P', 1)
# Compute magnitude of displacement
u_magnitude = sqrt(dot(u, u))
u_magnitude = project(u_magnitude, Uf)
p=plot(u_magnitude, title='Displacement magnitude')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(u[1], title='Displacement vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(u[0], title='Displacement horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()


ep=e_plas[0,0]+e_plas[1,1]
p=plot(ep, title='plastic strain, p')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

eq=2.*sqrt((e_plas[0,0]-e_plas[1,1])**2/4.+e_plas[0,1]**2)
p=plot(eq, title='plastic strain, q')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

# Plot stress
s = sig # all stress
V = FunctionSpace(mesh, 'P', 1)
s0 = s[0,0]
s1 = s[1,1]

p=plot(s0, title='Stress horizontal')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

p=plot(s1, title='Stress vertical')
# add a colorbar
plt.colorbar(p,format=fmt);
# Hold plot
plt.show()

print('min/max u:',
      u_magnitude.vector().get_local().min(),
      u_magnitude.vector().get_local().max())

# this integrates over the domain
energy = 0.5*inner(sig, epsilon(u))*dx
E = assemble(energy)	

print('energy:',
      E)

	  




