"""
Heat diffusion in a confined material
"""

# Importing the essentials
from fenics import *
import matplotlib.pyplot as plt
import numpy as np

# Geometry
Lx = 0.3	# width
Ly = 0.1	# height
nx = 30		# number of elements in the x-direction
ny = 10		# number of elements in the y-direction

# Material parameters for elasticity
nu = 0.2
E  = 1000000.0
lamda = Constant(E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu)))
mu    = Constant(E / (2.0 * (1.0 + nu)))

# Density and gravity
rho   = 1700.0
g     = 0.0 

# Thermal parameters
kcond = 0.25 # thermal conductivity
cp    = 800  # specific heat capacity
alpha = kcond/(rho*cp) # thermal diffusivity

# Preparing the mesh
mesh = RectangleMesh(Point(0,0), Point(Lx,Ly), nx, ny)
#plot(mesh, title="Mesh")
#plt.show()

# Defining elements
# For displacements
P2 = VectorElement("Lagrange", mesh.ufl_cell(), 2)
V  = FunctionSpace(mesh, P2)
# For the temperature
P1 = FiniteElement("Lagrange", mesh.ufl_cell(), 1)
Q  = FunctionSpace(mesh, P1)
# Make a mixed space
Z0 = P2 * P1
Z = FunctionSpace(mesh, Z0)
# Define trial and test functions
z = TrialFunction(Z)
(u,ut) = split(z)
zn = TestFunction(Z)
(v,vt) = split(zn)
# Tensor and vector dimensions
d = u.geometric_dimension()  # space dimension

# Define the boundaries
# left boundary
class LeftBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] < tol

# right boundary
class RightBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0]> Lx - tol	

class UpperBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[1] > Ly - tol

class LowerBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[1] < tol		

left_boundary = LeftBoundary()        
right_boundary = RightBoundary()
upper_boundary = UpperBoundary()
lower_boundary = UpperBoundary()

boundary_parts = MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
boundary_parts.set_all(0)
left_boundary.mark(boundary_parts, 1)
right_boundary.mark(boundary_parts, 2)
upper_boundary.mark(boundary_parts, 3)
lower_boundary.mark(boundary_parts, 4)
ds = ds(subdomain_data = boundary_parts)

# Define strain and stress
def epsilon(u):
    e = 0.5*(nabla_grad(u) + nabla_grad(u).T)
    return as_tensor([[e[0, 0], e[0, 1]],
                      [e[1, 0], e[1, 1]]])

# Stiffness tensor					  
def dsde():
    stiffness = np.zeros((d,d,d,d))
    for i in range(0,d):
#         stiffness[i,i,i,i] +=lamda
         for j in range(0,d):
             stiffness[i,i,j,j] +=lamda
             stiffness[i,j,i,j] +=mu
             stiffness[i,j,j,i] +=mu
    return as_tensor(stiffness)
	
# Stress tensor
def sigma(u):
#    print('dimensions:',
#         dsde().shape,
#         epsilon(u).shape)
#    sigma = np.tensordot(dsde(),epsilon(u),axes=([2,0],[3,1]))
    eps=epsilon(u)
    stiffness =dsde()
    sigma = as_tensor(stiffness[(i,j,k,l)]*eps[(k,l)],(i,j))	
    return as_tensor(sigma)
	
# Newton-Raphson matrix
def NR(u):
     stiffness =dsde()
     stf = as_tensor(stiffness[(i,j,k,l)]*nabla_grad(u)[(k,l)],(i,j))
     NR = inner(stf, nabla_grad(v))*dx
#     NR = inner(sigma(u), nabla_grad(v))*dx
     return NR

# Define variational problem
# Define gravity term
f = Constant((0, -rho*g))
# Assembling
a = inner(sigma(u), nabla_grad(v))*dx + alpha*inner(nabla_grad(ut),nabla_grad(vt))*dx
L = dot(f, v)*dx

# Fixing all displacements at the boundary
bcd = DirichletBC(Z.sub(0), Constant((0.0,0.0)), lower_boundary)	
bcu = DirichletBC(Z.sub(0), Constant((0.0,0.0)), upper_boundary)
bcl = DirichletBC(Z.sub(0), Constant((0.0,0.0)), left_boundary)
bcr = DirichletBC(Z.sub(0), Constant((0.0,0.0)), right_boundary)
# Fixing the temperatures
#bcdt= DirichletBC(Z.sub(1), Constant(0.0), lower_boundary)	
#bcut= DirichletBC(Z.sub(1), Constant(0.0), upper_boundary)
bclt= DirichletBC(Z.sub(1), Constant(100.0), left_boundary)
bcrt= DirichletBC(Z.sub(1), Constant(0.0), right_boundary)

bc = [bcd, bcu, bcl, bcr, bclt, bcrt]
# Compute solution
q = Function(Z)
solve(a == L, q, bc)

disp = Function(V)
T    = Function(Q)
(disp,T)=split(q)

fmt='%1.8f'

# Plot solution
p=plot(disp, title='Displacement', mode='displacement')
plt.colorbar(p,format=fmt)
font={'size':18}
plt.rc('font',**font)
# Hold plot
plt.show()

# Plot solution
p=plot(T, title='Temperature')
plt.colorbar(p,format=fmt)
font={'size':18}
plt.rc('font',**font)
# Hold plot
plt.show()

# this integrates over the domain
energy = 0.5*inner(sigma(disp), grad(disp))*dx
E = assemble(energy)	

print('energy:',
      E)
	  


