"""
Heat diffusion in a confined material
"""

# Importing the essentials
from fenics import *
#from fenicstools import interpolate_nonmatching_mesh_any
import numpy as np
import matplotlib.pyplot as plt
# Geometry
Lx = 0.3	# width
Ly = 0.1	# height
nx = 30		# number of elements in the x-direction
ny = 10		# number of elements in the y-direction

# Material parameters for elasticity
nu = 0.2
E  = 1000000.0
lamda = Constant(E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu)))
mu    = Constant(E / (2.0 * (1.0 + nu)))

# Density and gravity
rho   = 1700.0
g     = 0.0 

# Thermal parameters
kcond = 0.25*3600.     # thermal conductivity (time in hours)
cp    = 800            # specific heat capacity
alpha = kcond/(rho*cp) # thermal diffusivity (time in hours)

# Preparing the mesh
mesh = RectangleMesh(Point(0,0), Point(Lx,Ly), nx, ny)
#plot(mesh, title="Mesh")
#plt.show()
unitNormal = FacetNormal(mesh)

# Defining elements
# For displacements
P2 = VectorElement("Lagrange", mesh.ufl_cell(), 2, dim=2)
V  = FunctionSpace(mesh, P2)
# For the temperature
P1 = FiniteElement("Lagrange", mesh.ufl_cell(), 1)
Q  = FunctionSpace(mesh, P1)
# Make a mixed space
#Z0 = P2 * P1
#Z = FunctionSpace(mesh, Z0)
Z= FunctionSpace(mesh, MixedElement([P2, P1]))

# Define trial and test functions
z = TrialFunction(Z)
(u,ut) = split(z)
zn = TestFunction(Z)
(v,vt) = split(zn)
# Tensor and vector dimensions
d = u.geometric_dimension()  # space dimension

xc   = np.linspace(0,Lx,nx)
ures = np.zeros(nx)

# Define the boundaries
# left boundary
class LeftBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0] < tol

# right boundary
class RightBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[0]> Lx - tol	

class UpperBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[1] > Ly - tol

class LowerBoundary(SubDomain):
    def inside(self, x, on_boundary):
        tol = DOLFIN_EPS
        return on_boundary and x[1] < tol		

left_boundary = LeftBoundary()        
right_boundary = RightBoundary()
upper_boundary = UpperBoundary()
lower_boundary = UpperBoundary()

boundary_parts = MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
boundary_parts.set_all(0)
left_boundary.mark(boundary_parts, 1)
right_boundary.mark(boundary_parts, 2)
upper_boundary.mark(boundary_parts, 3)
lower_boundary.mark(boundary_parts, 4)
ds = ds(subdomain_data = boundary_parts)
ds1= ds(subdomain_data = left_boundary)
# Define strain and stress
def epsilon(u):
    e = 0.5*(nabla_grad(u) + nabla_grad(u).T)
    return as_tensor([[e[0, 0], e[0, 1]],
                      [e[1, 0], e[1, 1]]])

# Stiffness tensor					  
def dsde():
    stiffness = np.zeros((d,d,d,d))
    for i in range(0,d):
#         stiffness[i,i,i,i] +=lamda
         for j in range(0,d):
             stiffness[i,i,j,j] +=lamda
             stiffness[i,j,i,j] +=mu
             stiffness[i,j,j,i] +=mu
    return as_tensor(stiffness)
	
# Stress tensor
def sigma(u):
#    print('dimensions:',
#         dsde().shape,
#         epsilon(u).shape)
#    sigma = np.tensordot(dsde(),epsilon(u),axes=([2,0],[3,1]))
    eps=epsilon(u)
    stiffness =dsde()
    sigma = as_tensor(stiffness[(i,j,k,l)]*eps[(k,l)],(i,j))	
    return as_tensor(sigma)
	
# Newton-Raphson matrix
def NR(u):
     stiffness =dsde()
     stf = as_tensor(stiffness[(i,j,k,l)]*nabla_grad(u)[(k,l)],(i,j))
     NR = inner(stf, nabla_grad(v))*dx
#     NR = inner(sigma(u), nabla_grad(v))*dx
     return NR		
	 
# Extracting results
def extract(u,coor,sav,i):
     k =len(coor)
#     temp_l= project(sigma_w,Q)	 
     for j in range(k):
#        print(j,coor[j])
        sav[i,j] = u([coor[j],Ly/2.0])
     return 0 

# Fixing all displacements at the boundary
bcd = DirichletBC(Z.sub(0), Constant((0.0,0.0)), lower_boundary)	
bcu = DirichletBC(Z.sub(0), Constant((0.0,0.0)), upper_boundary)
bcl = DirichletBC(Z.sub(0), Constant((0.0,0.0)), left_boundary)
bcr = DirichletBC(Z.sub(0), Constant((0.0,0.0)), right_boundary)
# Fixing the temperatures
#bcdt= DirichletBC(Z.sub(1), Constant(0.0), lower_boundary)	
#bcut= DirichletBC(Z.sub(1), Constant(0.0), upper_boundary)
bclt= DirichletBC(Z.sub(1), Constant(100.0), left_boundary)
bcl0= DirichletBC(Z.sub(1), Constant(0.0), left_boundary)
bcrt= DirichletBC(Z.sub(1), Constant(0.0), right_boundary)

bc1= [bcd, bcu, bcl, bcr, bclt, bcrt]
bc0= [bcd, bcu, bcl, bcr, bcl0, bcrt]

# Concerning the time
Tf = 10.0           # final time in hours
num_steps = 100   	# number of time steps
dt = Tf / num_steps # time step size

# For the Newton - Raphson
prec=1e-9           # precision
niter_max=20        # max. number of iterations

# temperature distribution along the middle line
saved = np.zeros((num_steps,nx))
# heat inflow
Qin   = np.zeros(num_steps)
# stored energy
Qtot  = np.zeros(num_steps)

# useful variables
q = Function(Z, name = "Total values")
disp = Function(V, name = "Total displacement")
T    = Function(Q, name = "Total temperature")
(disp,T)=split(q)
Dq = Function(Z, name = "Current increment")
Du = Function(V, name = "Displacement increment")
DT = Function(Q, name = "Temperature increment")
(Du,DT)=split(Dq)
dq = Function(Z, name = "Iteration correction")
du = Function(V, name = "Displacement correction")
dT = Function(Q, name = "Temperature correction")
(du,dT)=split(dq)

# Define variational problem
# Define gravity term
f = Constant((0, -rho*g))
# Assembling
a = ...
L = ...

# Time-stepping
t = 0

#initializing
#q=interpolate([0,0,0],Z)
disp0 = interpolate(Constant((0,0)), Z.sub(0).collapse())
T0    = interpolate(Constant(0), Z.sub(1).collapse())
assign(q,[disp0,T0])

for n in range(num_steps):
# Update current time
    t += dt

# Starting the Newton-Raphson
    niter = 0   # initialize iterations
    nRes  = 1.  # initialize residue
    assign(Dq,[disp0,T0])
    bcs = bc1
    while nRes > prec and niter < niter_max:		
        solve(a == L, dq, bcs,solver_parameters={"linear_solver": "mumps"})
        Dq.assign(Dq+dq)	
        res = -inner(sigma(q.sub(0)+Dq.sub(0)), nabla_grad(v))*dx + dot(f, v)*dx+alpha*inner(nabla_grad(q.sub(1)+Dq.sub(1)), nabla_grad(vt))*dt*dx +  (Dq.sub(1)*vt)*dx		
        Res = assemble(res)
        for bc in bcs:
            bc.homogenize()
            bc.apply(Res)
        nRes = Res.norm("l2")
        ndu  = dq.vector().norm('l2')
        bcs = bc0		
        print("    Residual:", nRes , "Increment:", ndu ,"Iteration:", niter)
        niter += 1
    q.assign(q+dq)
    (tmp1,tmp2) = split(q)
    extract(tmp2,xc,saved,n)
#   place here the energy inflow Qin = ...
    ...
#   place here the heat energy total Qtot = ...		
    ...
	
	
fmt='%1.8f'

(disp,T)=split(q)

# Plot solution
p=plot(disp, title='Displacement', mode='displacement')
plt.colorbar(p,format=fmt)
font={'size':18}
plt.rc('font',**font)
# Hold plot
plt.show()

# Plot solution
p=plot(T, title='Temperature')
plt.colorbar(p,format=fmt)
font={'size':18}
plt.rc('font',**font)
# Hold plot
plt.show()

plt.plot(xc,saved[0],label='dt')
plt.plot(xc,saved[num_steps/2],label='mid')
plt.plot(xc,saved[num_steps-1],label='end')
plt.xlabel("$x$")
plt.ylabel("$E_{tot}$")
plt.legend(loc='upper left')
#plt.ylim(0,.0001)
font={'size':20}
plt.rc('font',**font)
plt.show()

time =np.linspace(0,Tf,num_steps)
plt.plot(time,(Qin-Qin[0])/1000,label='in')
plt.plot(time,(Qtot-Qtot[0])/1000,label='tot')
plt.xlabel("$t$ [h]")
plt.ylabel("$Q$ [kJ]")
plt.legend(loc='upper left')
font={'size':20}
plt.rc('font',**font)
plt.show()


# this integrates over the domain
energy = 0.5*inner(sigma(disp), grad(disp))*dx
E = assemble(energy)	

print('energy:',
      E)
	  


